@@ -46,6 +46,7 @@ from dkd import Content, Message
 from dkd import InstantMessage, SecureMessage, ReliableMessage
 from dkd import InstantMessageDelegate, ReliableMessageDelegate
 
+from .protocol import Command
 from .delegate import EntityDelegate, CipherKeyDelegate
 
 
@@ -112,6 +113,22 @@ class Transceiver(InstantMessageDelegate, ReliableMessageDelegate):
     #
     #  Transform
     #
+
+    def __overt_group(self, content: Content) -> Optional[ID]:
+        group = content.group
+        if group is None:
+            return None
+        group = self.barrack.identifier(group)
+        if group.is_broadcast:
+            # broadcast message is always overt
+            return group
+        if isinstance(content, Command):
+            # group command should be sent to each member directly, so
+            # don't expose group ID
+            return None
+        else:
+            return group
+
     def encrypt_message(self, msg: InstantMessage) -> SecureMessage:
         barrack = self.barrack
         sender = barrack.identifier(msg.envelope.sender)
@@ -129,11 +146,17 @@ class Transceiver(InstantMessageDelegate, ReliableMessageDelegate):
         #         if you don't want to share the symmetric key with other members,
         #         you could split it (set group ID into message content and
         #         set contact ID to the "receiver") before encrypting, this usually
-        #         for sending group command to robot assistant,
-        #         which cannot shared the symmetric key (msg key) with other members.
+        #         for sending group command to assistant robot, which should not
+        #         shared the symmetric key (group msg key) with other members.
 
         # 1. get symmetric key
-        password = self.__password(sender=sender, receiver=receiver)
+        group = self.__overt_group(content=msg.content)
+        if group is None:
+            # personal message or (group) command
+            password = self.__password(sender=sender, receiver=receiver)
+        else:
+            # group message (excludes group command)
+            password = self.__password(sender=sender, receiver=group)
         assert isinstance(password, dict), 'password error: %s' % password
 
         # check message delegate
@@ -149,6 +172,19 @@ class Transceiver(InstantMessageDelegate, ReliableMessageDelegate):
             # personal message (or split group message)
             assert receiver.is_user, 'unknown receiver type: %s' % receiver
             s_msg = msg.encrypt(password=password)
+
+        # overt group ID
+        if group is not None and receiver != group:
+            # NOTICE: this help the receiver knows the group ID
+            #         when the group message separated to multi-messages,
+            #         if don't want the others know you are the group members,
+            #         remove it.
+            s_msg.envelope.group = group
+
+        # NOTICE: copy content type to envelope
+        #         this help the intermediate nodes to recognize message type
+        s_msg.envelope.type = msg.content.type
+
         # OK
         return s_msg
 
@@ -297,7 +333,8 @@ class Transceiver(InstantMessageDelegate, ReliableMessageDelegate):
         sender = barrack.identifier(sender)
         receiver = barrack.identifier(receiver)
         if key is None:
-            password = None
+            # get key from cache
+            password = key_cache.cipher_key(sender=sender, receiver=receiver)
         else:
             # decrypt key data with the receiver's private key
             identifier = barrack.identifier(msg.envelope.receiver)
@@ -308,8 +345,6 @@ class Transceiver(InstantMessageDelegate, ReliableMessageDelegate):
                 raise AssertionError('failed to decrypt key in msg: %s' % msg)
             # deserialize it to symmetric key
             password = self.deserialize_key(key=plaintext, msg=msg)
-        # get/cache cipher key for reusing
-        password = key_cache.reuse_cipher_key(key=password, sender=sender, receiver=receiver)
         assert isinstance(password, dict), 'failed to decrypt key: %s -> %s' % (sender, receiver)
         return password
 
@@ -329,9 +364,24 @@ class Transceiver(InstantMessageDelegate, ReliableMessageDelegate):
         if plaintext is None:
             # raise ValueError('failed to decrypt data: %s, password: %s' % (data, password))
             return None
+        content = self.deserialize_content(data=plaintext, msg=msg)
+        assert content is not None, 'content error: %d' % len(plaintext)
+        # check and cache key for reuse
+        barrack = self.barrack
+        sender = barrack.identifier(msg.envelope.sender)
+        group = self.__overt_group(content=content)
+        if group is None:
+            receiver = barrack.identifier(msg.envelope.receiver)
+            # personal message or (group) command
+            # cache key with direction (sender -> receiver)
+            self.key_cache.cache_cipher_key(key=password, sender=sender, receiver=receiver)
+        else:
+            # group message (excludes group command)
+            # cache the key with direction (sender -> group)
+            self.key_cache.cache_cipher_key(key=password, sender=sender, receiver=group)
         # NOTICE: check attachment for File/Image/Audio/Video message content
         #         after deserialize content, this job should be do in subclass
-        return self.deserialize_content(data=plaintext, msg=msg)
+        return content
 
     def sign_data(self, data: bytes, sender: str, msg: SecureMessage) -> bytes:
         barrack = self.barrack
